#!/usr/pkg/bin/perl

use IO::Socket;
use IO::Socket::IP;
use MIME::Base64;
use Crypt::CBC;
use Getopt::Long;
use Fcntl qw(:flock);

sub usage {
    return "ugly_rnd_client.pl --rnd-host RND_HOST_SERVER [--rnd-port DEFAULT_9999] [--bits DEFAULT_128] [--get-all-bits] [--fill-up NUM] [--encrypt] [--key-file KEY_FILE_FOR_ENCRYPTION] [--key-id ID] [--lock-file FILE] [--output-file FILE] [--show-bits] [--timeout SECONDS] [--id ID] [--help]\n";
}

sub get_rndctl_stats {
    my($rndctl);
    my(@rndctl);
    $rndctl = `/sbin/rndctl -s`;
    @rndctl = split("\n",$rndctl);
    my($x);
    my($c);
    my($m);
    foreach $x (@rndctl) {
	if ($x =~ /^[ 	]*([0-9]*) bits currently stored in pool .max ([0-9]*)./) {
	    $c = $1;
	    $m = $2;
	    last;
	}
    }

    return($c,$m);
}

my($key_file);
my($want);
my($get_all_bits);
my($want_encryption);
my($rnd_host);
my($rnd_port) = 9999;
my($want_id);
my($help);
my($output_file);
my($fill_up) = 5;
my($show_bits);
my($lock_file);
my($timeout);
my($id);

GetOptions("key-file=s" => \$key_file,
	   "key-id=s" => \$want_id,
	   "bits=i" => \$want,
	   "get-all-bits" => \$get_all_bits,
	   "encrypt" => \$want_encryption,
	   "rnd-server|rnd-host=s" => \$rnd_host,
	   "rnd-port=i" => \$rnd_port,
	   "output-file=s" => \$output_file,
	   "fill-up=i" => \$fill_up,
	   "show-bits" => \$show_bits,
	   "lock-file=s" => \$lock_file,
	   "timeout=i" => \$timeout,
	   "id=s" => \$id,
	   "help" => \$help
    ) or die "Argument error:\n" . usage();

if (defined $help) {
    print usage();
    exit 0;
}

if (! defined $rnd_host) {
    print "Need host\n";
    print usage();
    exit 1;
}

die "Need key-id if key-file is defined\n" if ! defined $want_id && defined $key_file;

if (defined $timeout) {
    die "timeout must be greater than 0\n" if ($timeout <= 0);
}

my(%keys);
if (defined $key_file) {
    if ( -s "$key_file") {
        my($key);
        my($id);
	my($salt);
        open(KEYFILE,$key_file) || die "Could not open keyfile $key_file: $!";
        while (<KEYFILE>) {
            chomp;
            if (/^#/) {
                next;
            }
            if (/^$/) {
                next;
            }
            if (/^key:/i) {
                my($k,$v) = split(/:/);
                $key = $v;
            }
            if (/^id:/i) {
                my($k,$v) = split(/:/);
                $id = $v;
                undef $key;
                undef $salt;
            }
            if (/^salt:/i) {
                my($k,$v) = split(/:/);
                $salt = $v;
            }
            if (defined $id) {
                if (defined $key) {
                    $keys{$id}{'key'} = $key;
                }
                if (defined $salt) {
                    $keys{$id}{'salt'} = $salt;
                }
            }
        }
        close KEYFILE;
    }
}

if (defined $lock_file) {
    my($rr);
    open(LOCK,$lock_file);
    $rr = flock(LOCK,LOCK_EX | LOCK_NB);
    if ($rr == 0) {
	print "Already getting randomness...\n";
	exit;
    }
}

if (! defined $want) {
    if (! defined $get_all_bits) {
	$want = 128;
    } else {
	my($current);
	my($max);
	($current,$max) = get_rndctl_stats();
	if ($current + 0.0 < $max + 0.0) {
	    $want = ($max + 0.0) - ($current + 0.0);
	    print "Max bits: $max ; curent bits: $current ; want bits: $want\n";
	} else {
	    print "No need to get bits\n";
	    exit;
	}
    }
}

if (defined $want_encryption) {
    $want = $want * 2;
    # if ($want > 1024) {
    # 	$want = 1024;
    # }
}

print "Asking for $want bits\n";

my($handle);
$handle = IO::Socket::IP->new( Proto => "tcp",
				 PeerAddr => $rnd_host,
				 PeerPort => $rnd_port )
    or die "can't connect: $!\n";

$handle->autoflush(1);

my($msg);
my($smsg);

eval {
    local $SIG{ALRM} = sub { print "Timeout initial\n"; exit 1; };
    if (defined $timeout) {
	alarm $timeout;
    }

    $msg = <$handle>;
    print $msg;

    alarm 0;
};

eval {
    local $SIG{ALRM} = sub { print "Timeout ID\n"; exit 1; };
    if (defined $timeout) {
	alarm $timeout;
    }

    if (defined $id) {
	print $handle "ID $id\n";
	$msg = <$handle>;
	print $msg;
    }

    alarm 0;
};

eval {
    local $SIG{ALRM} = sub { print "Timeout ENCRYPT\n"; exit 1; };
    if (defined $timeout) {
	alarm $timeout;
    }

    if (defined $want_encryption) {
	print $handle "ENCRYPT $want_id\n";
	$msg = <$handle>;
	print $msg;
	my($has_encryption);
	if ($msg =~ /^202/) {
	    $has_encyption++;
	}
    }

    alarm 0;
};

_fill_up:

    eval {
	local $SIG{ALRM} = sub { print "Timeout GET want\n"; exit 1; };
	if (defined $timeout) {
	    alarm $timeout;
	}

	print $handle "GET $want\n";
	$msg = <$handle>;

	alarm 0;
};

$smsg = $msg;
if (! defined $show_bits) {
    $smsg =~ s/^([0-9]*).*/$1/;
}
print $smsg;
if ($msg =~ /^100/) {
    system('rndctl -s');
    my(@v);
    @v = split(/ - /,$msg);
    my($d);
    my($plaintext);
    $d = decode_base64($v[1]);
    if (defined $has_encyption) {
	my(%cargs);
	$cargs{'-key'} = $keys{$want_id}{'key'};
	$cargs{'-cipher'} = "Crypt::OpenSSL::AES";
	if (defined $keys{$want_id}{'salt'}) {
	    $cargs{'-salt'} = $keys{$want_id}{'salt'};
	}
	my($cipher) = Crypt::CBC->new(%cargs);

	# my($cipher) = Crypt::CBC->new(-key => $key,
	# 			      -cipher => "Crypt::OpenSSL::AES"
	#     );
	# my($cipher) = Crypt::CBC->new(-key => $key,
	# 			      -cipher => "Crypt::Rijndael"
	#     );
	$plaintext = $cipher->decrypt($d);
    } else {
	$plaintext = $d;
    }
    if (! defined $output_file) {
	open(RANDOMDEV,"> /dev/random") || die "Could not open random device: $!\n";
	my($wrote);
	$wrote = syswrite RANDOMDEV, $plaintext;
	if (defined $wrote) {
	    print "WROTE: $wrote\n";
	} else {
	    print "ERROR: $!\n";
	}
	close RANDOM;
	system('rndctl -s');
    } else {
	my($wrote);
	open(O,"> $output_file");
	$wrote = syswrite O, $plaintext;
	if (defined $wrote) {
	    print "WROTE: $wrote\n";
	} else {
	    print "ERROR: $!\n";
	}
	close O;
    }

    $fill_up--;
    if ($fill_up > 0) {
	if (defined $get_all_bits) {
	    my($current);
	    my($max);
	    ($current,$max) = get_rndctl_stats();
	    if ($current + 0.0 < $max + 0.0) {
		$want = ($max + 0.0) - ($current + 0.0);
		print "Filling up --> Max bits: $max ; curent bits: $current ; want bits: $want\n";
		goto _fill_up;
	    } else {
		print "Filling up --> No need to get bits\n";
		exit;
	    }
	} else {
	    print "Filling up --> Asking for $want bits\n";
	    goto _fill_up;
	}
    }
}

eval {
    local $SIG{ALRM} = sub { print "Timeout BYE\n"; exit 1; };
    if (defined $timeout) {
	alarm $timeout;
    }

    print $handle "BYE\n";
    $msg = <$handle>;
    print $msg;

    alarm 0;
};
